/**
 * A class representing a Binary Tree of elements (node-based).
 * Includes methods for printing out traversals of the tree.
 *
 * @author Joel Ross and CS261, Sp 2014
 *
 */
public class BinaryTree<E>
{
	//an inner class representing a node in the tree
	protected static class Node<E>
	{
		protected E data;
		protected Node<E> left, right;
		
		public Node(E data)
		{
			this.data = data;
			left = null;
			right = null;
		}
		
		public String toString()
		{
			return data.toString();
		}
	}
	
	protected Node<E> root; //the root
	
	/**
	 * Constructs a new BinaryTree with the given root and branches
	 * @param  data        the data for the root
	 * @param  leftBranch  the left subtree
	 * @param  rightBranch the right subtree
	 */
	public BinaryTree(E data, BinaryTree<E> leftBranch, BinaryTree<E> rightBranch)
	{
		root = new Node<E>(data);
		if(leftBranch != null)
			root.left = leftBranch.root;
		if(rightBranch != null)
			root.right = rightBranch.root;
	}
	
	/**
	 * Constructs a new BinaryTree with just this item as the root
	 * @param  data the data for the root
	 */
	public BinaryTree(E data)
	{
		this(data, null, null);
	}
	
	/**
	 * Constructs a new empty binary tree
	 */
	public BinaryTree()
	{
		this.root = null;
	}
	
	/**
	 * Constructs a new binary tree with the given node as the root
	 * @param  newRoot The node to be the new root
	 */
	protected BinaryTree(Node<E> newRoot)
	{
		this.root = newRoot;
	}

	/**
	 * Returns the left subtree of this tree
	 * @return the left subtree
	 */
	public BinaryTree<E> getLeftSubTree()
	{
		if(root != null)
			return new BinaryTree<E>(root.left);
		return new BinaryTree<E>();
	}

	/**
	 * Returns the right subtree of this tree
	 * @return the right subtree
	 */
	public BinaryTree<E> getRightSubTree()
	{
		if(root != null)
			return new BinaryTree<E>(root.right);
		return new BinaryTree<E>();
	}
	
	/**
	 * Returns the data in the root node
	 * @return The data in the root
	 */
	public E getData()
	{
		return root.data;
	}

	/**
	 * Prints the nodes in a pre-order traversal
	 */
	public void printPreOrderTraversal()
	{
		printPreOrderTraversal(root);
		System.out.println();

	}
	
	//helper method for the pre-order traversal; prints from the given local root
	private void printPreOrderTraversal(Node<E> localRoot)
	{
		System.out.print(localRoot+" ");
		if(localRoot.left != null)
			printPreOrderTraversal(localRoot.left);
		if(localRoot.right != null)
			printPreOrderTraversal(localRoot.right);
	}
	
	/**
	 * Prints the nodes in a in-order traversal
	 */
	public void printInOrderTraversal()
	{
		printInOrderTraversal(root);
		System.out.println();

	}

	//helper method for the in-order traversal; prints from the given local root
	private void printInOrderTraversal(Node<E> localRoot)
	{
		if(localRoot.left != null)
			printInOrderTraversal(localRoot.left);

		System.out.print(localRoot+" ");

		if(localRoot.right != null)
			printInOrderTraversal(localRoot.right);
	}
	
	/**
	 * Prints the nodes in a post-order traversal
	 */
	public void printPostOrderTraversal()
	{
		printPostOrderTraversal(root);
		System.out.println();

	}

	//helper method for the post-order traversal; prints from the given local root
	private void printPostOrderTraversal(Node<E> localRoot)
	{
		if(localRoot.left != null)
			printPostOrderTraversal(localRoot.left);

		if(localRoot.right != null)
			printPostOrderTraversal(localRoot.right);

		System.out.print(localRoot+" ");
	}
	
	//Returns a string that is a pre-order traversal
	public String toString()
	{
		return "["+toStringPreOrder(root)+"]";
	}
	
	/**
	 * Helper method that returns a string in pre-order traversal
	 * @param  localRoot subtree to get string of
	 * @return           String representing this subtree
	 */
	private String toStringPreOrder(Node<E> localRoot)
	{
		String toReturn = localRoot.toString();

		if(localRoot.left != null)
			toReturn += ", "+ toStringPreOrder(localRoot.left);
		else
			toReturn += ", null";
		if(localRoot.right != null)
			toReturn += ", "+ toStringPreOrder(localRoot.right);
		else
			toReturn += ", null";
			
		return toReturn;
	}
	
	/**
	 * Prints out a graphical representation of the tree
	 */
	public void printTree()
	{
		System.out.println(getTreeString(root, ""));
	}
	
	private static final String SPACE_DELIM = "  "; //constant for spacing
	
	/**
	 * Returns a String representation of the given subtree
	 * @param  localRoot    root of subtree
	 * @param  currentSpace current indentation level
	 * @return              string representation of subtree
	 */
	private String getTreeString(Node<E> localRoot, String currentSpace)
	{
		String toReturn = "";
		
		if(localRoot.right != null)
			toReturn += getTreeString(localRoot.right, currentSpace+SPACE_DELIM);
		//else
		//	toReturn += currentSpace + SPACE_DELIM + "null\n";

		toReturn += currentSpace + localRoot.data + "\n";
		
		if(localRoot.left != null)
			toReturn += getTreeString(localRoot.left, currentSpace+SPACE_DELIM);
		//else 
		//	toReturn += currentSpace + SPACE_DELIM + "null\n";

		return toReturn;
	}
}
